package src.BAALL

import akka.actor.ActorDSL._
import akka.actor.{ActorLogging, ReceiveTimeout, ActorRef}
import org.semanticweb.owlapi.reasoner.impl.OWLClassNode
import scala.Some
import scala.collection.parallel.mutable.ParHashMap
import scala.concurrent.duration._
import src.ship._

/**
 * Created with IntelliJ IDEA.
 * User: dodi
 * Date: 11.09.12
 * Time: 10:29
 * propagation of updates in the real world to the ontology. registeredObjects is a hashmap that stores all known
 * lknxobjects together with their current values. the hashmap is used to compute the diffs
 * it is currently not supported to add objects at runtime, even though this should be possible in principle
 */

case class OntUpdater(lknx: Linknx, shipTool: ActorRef, registeredObjects: ParHashMap[String, LKNXObject])
  extends JavaLogging with ConceptParser {
  val backMap = Map(
    "Light" -> Map("off" -> "Off", "on" -> "On"),
    "Door" -> Map("off" -> "Closed", "on" -> "Open"),
    "DimLight" -> Map(
      "67" -> "DimDark",
      "127" -> "DimHalf",
      "190" -> "DimBright",
      "255" -> "DimOn",
      "0" -> "DimOff")
  )
  var ontology: Ontology = null
  var modifiedObjects: List[LKNXObject] = List[LKNXObject]()
  var buttonHistory: BAALLButtons = new BAALLButtons
  var buttonHistoryChanged = false

  /**
   * computes the ABox representation for a given LKNXObject. This is done by checking whether one of its possible types
   * occurs in the mapping defined in backMap
   * @param x LKNXObject to be translated
   * @return an ABoxFormula
   */
  def translate2ABox(x: LKNXObject): List[ABoxFormula] = {
    log.info("translate2Abox " + x + " " + ontology)
    var result: List[ABoxFormula] = Nil
    if (ontology != null) {
      val pm = Builtin.prefixManager
      x match {
        case ValObject(name, status) =>
          log.info("check change on " + ValObject(name, status))
          val types = ontology.getReasoner.getTypes(ontology.getManager.getOWLDataFactory.getOWLNamedIndividual(":" + name, pm), false)
          log.info("types.size = " + types.getNodes.size())
          val iter = types.getNodes.iterator()
          while (iter.hasNext) {
            val obj = iter.next().asInstanceOf[OWLClassNode]
            backMap.get(obj.getRepresentativeElement.getIRI.getFragment) match {
              case Some(m) =>
                log.info("found backmap " + m)
                //@todo to be done
                m.get(status.toString) match {
                  case Some(s) =>
                    log.info("Query1 " + ConceptConstant(s).mkAssertion(Const(name)))
                    if (ontology.query(ConceptConstant(s).mkAssertion(Const(name))).isEmpty) {
                      result = List(ConceptConstant(s).mkAssertion(Const(name)),
                        NamedRole("hasauthority").roleAssert(Const(name), Const("auser")))
                    }
                    log.info("Result = " + result)
                  case None =>
                }
              case None =>
            }
          }
        case PosObject(name, pos) =>
          if (ontology.query(ConceptAssertion(Const(pos), ConceptConstant("Position"))).isEmpty)
            log.info("Unknown object position " + pos + " for object " + name)
          else result = List(RoleAssertion(Const(name), Const(pos), NamedRole("at")))
        case SwitchObject(name, status) =>
          log.info("check change on " + SwitchObject(name, status))
          val types = ontology.getReasoner.getTypes(ontology.getManager.getOWLDataFactory.getOWLNamedIndividual(":" + name, pm), false)
          log.info("types.size = " + types.getNodes.size())
          val iter = types.getNodes.iterator()
          while (iter.hasNext) {
            val obj = iter.next().asInstanceOf[OWLClassNode]
            backMap.get(obj.getRepresentativeElement.getIRI.getFragment) match {
              case Some(m) =>
                log.info("found map " + m + " result is " + m.get(if (status) "on" else "off"))
                val assertion = ConceptAssertion(Const(name), ConceptConstant(m.get(if (status) "on" else "off").get))
                //log.info("Query2 " + assertion + " = " + ontology.query(assertion))
                if (ontology.query(assertion).isEmpty) {
                  result = List(assertion, NamedRole("hasauthority").roleAssert(Const(name), Const("auser")))
                }
                log.info("Result = " + result)
              case None =>
            }
          }
        case v@StringObject(name,value) =>
          if (isLiteral(value)) {
            result = List(RoleAssertion(Const(name), Literal(value), DataRole("hasStringValue")))
            log.info("Result = " + result)
          } else
          log.info("Ignoring Baall StringObject " + v)
        case v => log.info("Ignoring Baall Object " + v)
      }
      result
    }
    else {
      log.warning("ontology is null! ")
      result
    }
  }

  /**
   * sends the modified objects and clears the list of modified updates
   */
  def sendUpdates() {
    this.synchronized {
      if (!modifiedObjects.isEmpty || buttonHistoryChanged) {
        //println  ("modified objects: " + modifiedObjects)
        log.info("modified objects: " + modifiedObjects)
        //warning done for debugging
        var aboxupdates = List[ABoxFormula]()
        var aboxdeletes = List[Const]()
        for (obj <- modifiedObjects) {
          val res = translate2ABox(obj)
          log.info("res is " + res)
          if (res != Nil && ontology != null && ontology.query(res).isEmpty)
            aboxupdates = res ++ aboxupdates
          //println("registered object is now "+(registeredObjects.get(objname).get).toString)
        }

        if (buttonHistoryChanged) {
          (buttonHistory.getButtons :\ (None:Option[BAALLButtonEvent])) { (b,last) =>
            aboxupdates = aboxupdates ++ b.toOntology(ontology,last)
            if (ontology!=null) {
              ontology.query("(?be,:"+b.target+"):button_target") map {
                subst =>
                  val x = subst(Var("be")).asInstanceOf[Const]
                  println("subst "+x)
                  aboxdeletes = List(x) ++ aboxdeletes
              }
            }
            Some(b)
          }
          if (ontology!=null) {
            aboxupdates = aboxupdates ++ {
              ontology.query("(?x,?y):previous_buttonevent") flatMap { subst =>
              val res1 = subst(Var("x")).asInstanceOf[Const]
              val res2 = subst(Var("y")).asInstanceOf[Const]
              buttonHistory.getPreviousButton(res1.name) match {
                case Some(BAALLButtonEvent(id,_,_,_)) if id==res2.name => Nil
                case _ => List(NotRole(NamedRole("previous_buttonevent")).roleAssert(res1,res2))
              }
              //println("Subst "+subst+" var ?x --> "+res)
              }
            }
          }

          buttonHistoryChanged = false
        }

        if (!aboxupdates.isEmpty) {
          log.info("send updates " + aboxupdates + "   Deleting: " + aboxdeletes)

          SHIPTool.ref ! Update(aboxupdates, aboxdeletes)
        }
        modifiedObjects = List[LKNXObject]()
      }
    }
  }
}

class TimerActor(u: OntUpdater, timeout: Int, isYamamoto: Boolean) extends ActWithStash with ActorLogging {
  context.setReceiveTimeout(timeout.millisecond)

  var reset: Boolean = true

  def readRemoteStatus() = {
    val objects = u.lknx.readAllObjects()
    //println("ONTUPDATER read objects " + objects)
    this.synchronized {
      for (obj <- objects) {
        u.registeredObjects.get(u.lknx.getName(obj)) match {
          case Some(o) => if (reset || o != obj) {
            log.info("compare " + o + " with " + obj)
            u.registeredObjects.put(u.lknx.getName(obj), obj)
            u.modifiedObjects = obj :: u.modifiedObjects
          }
          case None => log.warning("object " + obj + " not registered in " + u.registeredObjects)
            u.registeredObjects.put(u.lknx.getName(obj), obj)
            u.modifiedObjects = obj :: u.modifiedObjects
        }
      }
    }
    u.sendUpdates()
  }

  override def receive = {
    case src.ship.Stop() =>
      log.warning("Timeractor received Stop")
      reset = true
      u.ontology = null
    case ReceiveTimeout =>
      //println("HALLO!! u = "+u.ontology)
      if (true || !isYamamoto) {
        val x = u.lknx.readLastButton()
        //println("readlastbutton "+x)
        x match {
          case bs@(_ :: _) =>
            bs foreach u.buttonHistory.insertButton
            u.buttonHistoryChanged = true
            //println("ONTUPDATER " + u.buttonHistory.getButtons)
          case _ =>
        }
      }
      if (u.ontology != null) {
        readRemoteStatus()
        if (reset) reset = false
      }
  }
}
