package src.navi

import java.io.{StringWriter, PrintWriter, File}
import org.antlr.runtime.{CommonTokenStream, ANTLRFileStream}
import scala.collection.JavaConversions
import scala.sys.process._
import scala.xml.PrettyPrinter
import src.BAALL._

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 17.09.2012
 * Time: 14:49
 */
object NavigationCFG2Ontology {
  def writeFile(f: File)(p: java.io.PrintWriter => Unit) {
    val pw = new PrintWriter(f)
    try {
      p(pw)
    } finally {
      pw.close()
    }
  }

  def getWayPoints(parser: MapParser): List[WayPoint] =
    JavaConversions.asScalaSet(parser.getMap.getEdges.keySet()).toList.sortWith((wp1, wp2) => wp1.getName <= wp2.getName)

  def getTargets(wp: WayPoint): List[WPEdge] =
    JavaConversions.collectionAsScalaIterable(wp.getEdges).toList.sortWith((e1, e2) => e1.getTarget <= e2.getTarget)

  def todot(parser: MapParser): String = {
    val es = getWayPoints(parser)
    val sbnodes = StringBuilder.newBuilder
    val sbedges = StringBuilder.newBuilder
    sbnodes.append("digraph {\n node [style=filled,fillcolor=red]\n")
    es.map(wp => {
      sbnodes.append("\"" + wp.getName + "\" [label=\"" + wp.getName + "\"]\n")
      getTargets(wp).map(edge =>
        sbedges.append("\"" + wp.getName + "\" -> \"" + edge.getTarget + "\"\n"))
    })
    sbedges.append("}\n")
    sbnodes.append(sbedges.toString())
    sbnodes.toString()
  }

  def toDests(parser: MapParser): DrivingDestinations = {
    val es = getWayPoints(parser)
    val ds = es.map(wp => {
      val ns = getTargets(wp).map(e =>
        NavigationGraphEdge("oneway", if (e.getType.equalsIgnoreCase("L")) "line" else "reverseOnlyLine",
          PosPrefix.add(e.getTarget))
      )
      DrivingDestination(PosPrefix.add(wp.getName), wp.getX.toString, wp.getY.toString, ns)
    })
    DrivingDestinations(ds)
  }

  def toXml(dests: DrivingDestinations): String = {
    val pp = new PrettyPrinter(100, 2)
    val sb = StringBuilder.newBuilder
    pp.format(dests.toXML, sb)
    sb.toString()
  }

  def toLatex(dests: DrivingDestinations): String = {
    var result = new StringWriter()
    dests.destinations map {
      d =>
        result.write(" \\position{(" + d.x + "," + d.y + ")}{" + d.label.replaceAll("_", "") + "};\n")
    }
    dests.destinations map {
      d =>
        d.edges map {
          e =>
            result.write(" \\edge{" + d.label.replaceAll("_", "") + "}{" + e.targetNode.replaceAll("_", "") + "};\n")
        }
    }
    val locconflicts = dests.conflictingNodeEdges(1000)
    locconflicts.nodeEdgeConflictMap map {
      entry => {
        val node = entry._1
        entry._2 map (edge => {
          val segstart = edge._1
          val segend = edge._2
          result.write("  \\conflictNodeEdge{" + node.replaceAll("_", "") + "}{" + segstart.replaceAll("_", "") + "}{" + segend.replaceAll("_", "") + "};\n")
        })
      }
    }
    locconflicts.nodeNodeConflictMap map {
      entry => {
        val node = entry._1
        entry._2 map (othernode => {
          result.write("  \\conflictNodeNode{" + node.replaceAll("_", "") + "}{" + othernode.replaceAll("_", "") + "};\n")
        })
      }
    }
    result.toString
  }

  def toOnt(dests: DrivingDestinations): String = {
    val es = dests.destinations
    def findWP(s: String): DrivingDestination = dests.findNode(s).get
    val upperRight = findWP("pos_187911111")
    val lowerLeft = findWP("pos_296758459")
    val bathroomDoor = findWP("pos_1330609446644")
    val sb = StringBuilder.newBuilder
    es.map(wp => {
      val pos = wp.label
      val x = wp.x.toDouble
      sb.append(pos + ":Position\n")
      sb.append("(" + pos + ",nowc):position_status\n")
      sb.append("(" + pos + "," + pos + "):pickup_position\n")
      if (x <= lowerLeft.x.toDouble)
        sb.append("(" + pos + ",kitchenliving):isinarea\n")
      else if (x >= upperRight.x.toDouble)
        sb.append("(" + pos + ",bedroom):isinarea\n")
      else if (wp.y.toDouble <= bathroomDoor.y.toDouble)
        sb.append("(" + pos + ",bathroom):isinarea\n")
      else sb.append("(" + pos + ",corridor):isinarea\n")
      if (wp == upperRight || wp == lowerLeft)
        sb.append("(" + pos + ",corridor):isinarea\n")
      wp.edges.map(edge =>
        sb.append("(" + pos + "," + edge.targetNode + "):neighbour\n"))
    })
    sb.toString()
  }

  def parseRouteGraph(f: File): MapParser = {
    val input: ANTLRFileStream = new ANTLRFileStream(f.getAbsolutePath)
    val lex: MapLexer = new MapLexer(input)
    val tokens: CommonTokenStream = new CommonTokenStream(lex)
    val parser: MapParser = new MapParser(tokens)
    parser.map() // this is doing the parsing !!
    parser
  }

  def main(args: Array[String]) {
    val cfgFile = "src/navi/navigation-graph-baall.cfg"
    val dotFile = "src/navi/navigation-graph-baall.dot"
    val parser: MapParser = parseRouteGraph(new File(cfgFile))
    val dests: DrivingDestinations = toDests(parser)
    writeFile(new File("config/RouteGraph.xml"))(
      _.write(toXml(dests)))
    writeFile(new File("doc/RouteGraph.tex"))(
      _.write(toLatex(dests)))
    writeFile(new File("examples/baall/RouteGraph.ont"))(
      _.write(toOnt(dests)))
    writeFile(new File(dotFile))(
      _.write(todot(parser)))
    println(toDests(parser).conflictingNodeEdges(1000).edgeConflictsDeleted)
    val dotbinary = "/usr/local/bin/dot"
    println("dotbinary = " + dotbinary)
    Seq(dotbinary, "-Tpdf", "-GLR", "-Kfdp",
      dotFile, "-osrc/navi/navigation-graph-baall.pdf").!!
  }
}
