package src.ship

import org.semanticweb.owlapi.model.PrefixManager
import org.semanticweb.owlapi.util.DefaultPrefixManager
import org.semanticweb.owlapi.vocab.OWLFacet
import scala.util.parsing.combinator.JavaTokenParsers

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 13.06.2012
 * Time: 17:02
 */

/**
 * Concepts and Roles (ALCO@ with nominals resp. ALCO+)
 */
object Builtin {
  def upCase1(s: String): String = s.head.toUpper + s.tail

  val keys: List[String] = {
    val ints = List("long", "int", "short", "byte")
    val posOrNeg = List("positive", "negative")
    val is = "integer"
    val ds = List(is, "boolean", "string", "decimal",
      "float", "double", "time", "date", "dateTime", "duration")

    def cat(s1: String)(s2: String): String = s1 + upCase1(s2)
    val unsInts = ints.map(cat("unsigned"))
    val pnIs = posOrNeg.map(cat(_)(is))
    val nPnIs = pnIs.map(cat("non"))
    val ads = ds ++ ints ++ unsInts ++ pnIs ++ nPnIs
    // println(ads)
    ads
  }
  val namespace: String = "http://www.semanticweb.org/ontologies/2012/6/ship.owl#"
  val prefixManager: PrefixManager = new DefaultPrefixManager(namespace)
}

trait ConceptParser extends JavaTokenParsers {
  def string: Parser[String] = "[a-zA-Z][a-zA-Z0-9_]*".r

  def quotedstring: Parser[String] = opt("\"[^\"]*\"".r) ^^ {
    case Some(s) => if (!s.isEmpty) s.substring(1, s.length - 1) else ""
    case None => ""
  }

  def literal: Parser[String] = floatingPointNumber | stringLiteral

  def isLiteral(l: String): Boolean =
    parseAll(literal, l) match {
      case Success(_, _) => true
      case _ => false
    }

  def concept: Parser[Concept] = andConcept ~ opt("+" ~ concept) ^^ {
    case f ~ Some(_ ~ g) => OrConcept(f, g)
    case f ~ None => f
  }

  def andConcept: Parser[Concept] = primConcept ~ opt("&" ~ andConcept) ^^ {
    case f ~ Some(_ ~ g) => AndConcept(f, g)
    case f ~ None => f
  }

  def primConcept: Parser[Concept] = notConcept | cardRestriction | forallRole | existsRole |
    nominalConcept | atConcept |
    "(" ~ concept ~ ")" ^^ {
      case "(" ~ c ~ ")" => c
    } |
    constConcept | dataConcept

  def facet: Parser[OWLFacet] = OWLFacet.values() map {
    m => literal(m.getSymbolicForm) ^^^ m
  } reduceLeft (_ | _)

  def restLiteral: Parser[Literal] = (literal | string) ^^ Literal

  def dataRestriction: Parser[Restriction] =
    facet ~ restLiteral ^^ {
      case f ~ l => Restriction(f, l)
    }

  def dataRestrictions: Parser[List[Restriction]] = "[" ~ repsep(dataRestriction, ",") ~ "]" ^^ {
    case "[" ~ l ~ "]" => l
  }

  def oneOfLits: Parser[List[Literal]] = "{" ~ repsep(literal, ",") ~ "}" ^^ {
    case "{" ~ l ~ "}" => l map Literal
  }

  def constConcept: Parser[Concept] = string ^? {
    case "T" => UniversalConcept
    case s if !Builtin.keys.contains(s) => ConceptConstant(s)
  }

  def dataConcept: Parser[Concept] = string ~ opt(dataRestrictions) ~ opt(oneOfLits) ^^ {
    case s ~ Some(l) ~ None => DataConcept(s, l, Nil)
    case s ~ None ~ Some(l) => DataConcept(s, Nil, l)
    case s ~ None ~ None if Builtin.keys.contains(s) => DataConcept(s, Nil, Nil)
  }

  def nominalConcept: Parser[Concept] = "{" ~ string ~ "}" ^^ {
    case "{" ~ s ~ "}" => Nominal(Var(s))
  }

  def atConcept: Parser[Concept] = "@" ~ string ~ "." ~ concept ^^ {
    case "@" ~ s ~ "." ~ c => At(Var(s), c)
  }

  def notConcept: Parser[Concept] = "~" ~ primConcept ^^ {
    case "~" ~ c => c.Negate
  }

  def forallRole: Parser[Concept] = "all" ~ role ~ "." ~ concept ^^ {
    case "all" ~ r ~ "." ~ c => (c, r) match {
      case (DataConcept(_, _, _), NamedRole(n)) => ForallRole(DataRole(n), c)
      case _ => ForallRole(r, c)
    }
  }

  def existsRole: Parser[Concept] = "ex" ~ role ~ "." ~ concept ^^ {
    case "ex" ~ r ~ "." ~ c => (c, r) match {
      case (DataConcept(_, _, _), NamedRole(n)) => ExistsRole(DataRole(n), c)
      case _ => ExistsRole(r, c)
    }
  }

  def toInt(n: String): Int = try Integer.parseInt(n) catch {
    case _: NumberFormatException => -1
  }

  def cardRestriction: Parser[Concept] = (">=" | "<=" | "=") ~ floatingPointNumber ~ role ~ "." ~ concept ^? {
    case s ~ n ~ r ~ _ ~ c if toInt(n) >= 0 =>
      val i = toInt(n)
      s match {
        case ">=" => MinRestriction(r, i, c)
        case "<=" => MaxRestriction(r, i, c)
        case "=" => AndConcept(MinRestriction(r, i, c), MaxRestriction(r, i, c))
      }
  }

  // SameAs and Different are predefined roles !!
  def role: Parser[Role] = andRole ~ opt("+" ~ role) ^^ {
    case f ~ Some(_ ~ g) => OrRole(f, g)
    case f ~ None => f
  }

  def andRole: Parser[Role] = primRole ~ opt("&" ~ andRole) ^^ {
    case f ~ Some(_ ~ g) => AndRole(f, g)
    case f ~ None => f
  }

  def primRole = notRole | inverseRole | nominalRole |
    "(" ~ role ~ ")" ^^ {
      case "(" ~ r ~ ")" => r
    } | namedRole

  def namedRole: Parser[Role] = string ^^ {
    case s => NamedRole(s)
  }

  def nominalRole: Parser[Role] = "{" ~ "(" ~ string ~ "," ~ string ~ ")" ~ "}" ^^ {
    case "{" ~ "(" ~ s1 ~ "," ~ s2 ~ ")" ~ "}" => NominalRole(s1, s2)
  }

  def inverseRole: Parser[Role] = "inv" ~ "(" ~ role ~ ")" ^^ {
    case "inv" ~ "(" ~ r ~ ")" => InverseRole(r)
  }

  def notRole: Parser[Role] = "~" ~ role ^^ {
    case "~" ~ r => NotRole(r)
  }
}

case class Restriction(facet: OWLFacet, lit: Literal) {
  override def toString = facet.getSymbolicForm + " " + lit
}

sealed trait Concept extends ConceptOrRole with GroundableObject[Concept] {

  def mkAssertion(i: VarOrConstant): ConceptAssertion = ConceptAssertion(i, this)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = this

  def isNegationOf(c: Concept): Boolean = {
    this match {
      case UniversalConcept => c match {
        case NotConcept(UniversalConcept) => true
        case _ => false
      }
      case ConceptConstant(n) => c match {
        case NotConcept(ConceptConstant(n1)) => n == n1
        case _ => false
      }
      case ConceptVariable(n) => c match {
        case NotConcept(ConceptVariable(n1)) => n == n1
        case _ => false
      }
      case Nominal(n) => c match {
        case NotConcept(Nominal(n1)) => n == n1
        case _ => false
      }
      case NotConcept(sub) => sub == c
      case _ => false
    }
  }

  def contains(c: Concept): Boolean = {
    if (this == c) true
    else this match {
      case AndConcept(c1, c2) => c1.contains(c) | c2.contains(c)
      case OrConcept(c1, c2) => c1.contains(c) | c2.contains(c)
      case NotConcept(c1) => c1.contains(c)
      case ForallRole(r, c1) => c1.contains(c)
      case ExistsRole(r, c1) => c1.contains(c)
      case MinRestriction(r, i, c1) => c1.contains(c)
      case MaxRestriction(r, i, c1) => c1.contains(c)
      case At(_, c1) => c1.contains(c)
      case _ => false
    }
  }

  def isPrefixConcept: Boolean = {
    this match {
      case ForallRole(_, _) => true
      case ExistsRole(_, _) => true
      case MinRestriction(_, _, _) => true
      case MaxRestriction(_, _, _) => true
      case At(_, _) => true
      case _ => false
    }
  }

  def Negate: Concept = NotConcept(this)
}

case object UniversalConcept extends Concept {
  override def toString = "T" // \u22A4"

  override def isLiteral = true

  override def isAtomic = true
}

case class ConceptConstant(name: String) extends Concept {
  var namespace: String = Builtin.namespace

  override def toString = name

  override def isLiteral = true

  override def isAtomic = true

  override def getConcepts: Set[ConceptConstant] = Set(this)

  def setNamespace(s: String): ConceptConstant = {
    namespace = s
    this
  }
}

case class ConceptVariable(name: String) extends Concept {
  override def toString = name.toUpperCase

  override def isAtomic = true

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = ConceptConstant(name)
}

case class DataConcept(name: String, restrs: List[Restriction], oneOf: List[Literal]) extends Concept {
  override def toString = name + (if (restrs.isEmpty) "" else restrs.mkString("[", ", ", "]")) +
    (if (oneOf.isEmpty) "" else oneOf.mkString("{", ", ", "}"))

  override def isLiteral = true

  override def isAtomic = true
}

case class NotConcept(sub: Concept) extends Concept {
  override def freeVars = sub.freeVars

  override def toString = sub match {
    case UniversalConcept => "\u22A5"
    case _ => // "\u00AC"
      "~" + (if (sub.isLiteral) " " + sub else "(" + sub + ")")
  }

  override def isLiteral = sub.isAtomic

  override def Negate = sub

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = NotConcept(sub.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = sub.getConcepts

  override def getRoles: Set[NamedRole] = sub.getRoles

  override def getIndividuals: Set[String] = sub.getIndividuals
}

case class AndConcept(left: Concept, right: Concept) extends Concept {
  override def toString = "(" +
    ((left, right) match {
      case (MinRestriction(r1, n1, c1), MaxRestriction(r2, n2, c2)) if r1 == r2 && n1 == n2 && c1 == c2 =>
        "= " + n1 + " " + r1 + "." + (if (c1.isLiteral) c1 else "(" + c1 + ")")
      case (MaxRestriction(r1, n1, c1), MinRestriction(r2, n2, c2)) if r1 == r2 && n1 == n2 && c1 == c2 =>
        "= " + n1 + " " + r1 + "." + (if (c1.isLiteral) c1 else "(" + c1 + ")")
      case _ =>
        (if (left.isPrefixConcept) "(" + left + ")" else left) +
          " & " + right
    }) + ")"


  override def freeVars = left.freeVars ++ right.freeVars

  override def Negate = OrConcept(left.Negate, right.Negate)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = AndConcept(left.groundBy(exceptions), right.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = left.getConcepts ++ right.getConcepts

  override def getRoles: Set[NamedRole] = left.getRoles ++ right.getRoles

  override def getIndividuals: Set[String] = left.getIndividuals ++ right.getIndividuals
}

case class OrConcept(left: Concept, right: Concept) extends Concept {
  override def toString = "(" +
    (if (left.isPrefixConcept) "(" + left + ")" else left) +
    " + " + right + ")"

  override def freeVars = left.freeVars ++ right.freeVars

  override def Negate = AndConcept(left.Negate, right.Negate)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = OrConcept(left.groundBy(exceptions), right.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = left.getConcepts ++ right.getConcepts

  override def getRoles: Set[NamedRole] = left.getRoles ++ right.getRoles

  override def getIndividuals: Set[String] = left.getIndividuals ++ right.getIndividuals
}

case class ForallRole(role: Role, sub: Concept) extends Concept {
  override def toString = // "\u2200"
    "all " + role + "." + (if (sub.isLiteral) sub else "(" + sub + ")")

  override def freeVars = role.freeVars ++ sub.freeVars

  override def Negate = ExistsRole(role, sub.Negate)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = ForallRole(role.groundBy(exceptions), sub.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = role.getConcepts ++ sub.getConcepts

  override def getRoles: Set[NamedRole] = role.getRoles ++ sub.getRoles

  override def getIndividuals: Set[String] = role.getIndividuals ++ sub.getIndividuals
}

case class ExistsRole(role: Role, sub: Concept) extends Concept {
  override def toString = // "\u2203"
    "ex " + role + "." + (if (sub.isLiteral) sub else "(" + sub + ")")

  override def freeVars = role.freeVars ++ sub.freeVars

  override def Negate = ForallRole(role, sub.Negate)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = ExistsRole(role.groundBy(exceptions), sub.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = role.getConcepts ++ sub.getConcepts

  override def getRoles: Set[NamedRole] = role.getRoles ++ sub.getRoles

  override def getIndividuals: Set[String] = role.getIndividuals ++ sub.getIndividuals
}

case class MinRestriction(role: Role, card: Int, sub: Concept) extends Concept {
  assert(card >= 0)

  override def toString = // "\u2265"
    ">= " + card + " " + role + "." + (if (sub.isLiteral) sub else "(" + sub + ")")

  override def freeVars = role.freeVars ++ sub.freeVars

  override def Negate = card match {
    case 0 => ForallRole(role, sub.Negate)
    case _ => MaxRestriction(role, card - 1, sub)
  }

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = MinRestriction(role.groundBy(exceptions), card, sub.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = role.getConcepts ++ sub.getConcepts

  override def getRoles: Set[NamedRole] = role.getRoles ++ sub.getRoles

  override def getIndividuals: Set[String] = role.getIndividuals ++ sub.getIndividuals
}

case class MaxRestriction(role: Role, card: Int, sub: Concept) extends Concept {
  assert(card >= 0)

  override def toString = // "\u2264"
    "<= " + card + " " + role + "." + (if (sub.isLiteral) sub else "(" + sub + ")")

  override def freeVars = role.freeVars ++ sub.freeVars

  override def Negate = MinRestriction(role, card + 1, sub)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = MaxRestriction(role.groundBy(exceptions), card, sub.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = role.getConcepts ++ sub.getConcepts

  override def getRoles: Set[NamedRole] = role.getRoles ++ sub.getRoles

  override def getIndividuals: Set[String] = role.getIndividuals ++ sub.getIndividuals
}

case class Nominal(v: VarOrConstant) extends Concept {
  override def toString = "{" + v + "}"

  override def freeVars = v match {
    case Var(n) => Set[ObjectLevelVariable](Var(n))
    case _ => Set.empty[ObjectLevelVariable]
  }

  override def isAtomic = true

  override def isLiteral = true

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = v match {
    case Const(n) => this
    case x@Var(n) if exceptions.contains(x) => this
    case Var(n) => Nominal(Const(n))
  }

  override def getIndividuals: Set[String] = v match {
    case Const(n) => Set(n)
    case _ => Set.empty
  }
}

case class At(v: VarOrConstant, sub: Concept) extends Concept {
  override def toString = "@" + v + "." + (if (sub.isLiteral) sub else "(" + sub + ")")

  override def freeVars = (v match {
    case Var(_) => Set[ObjectLevelVariable](v)
    case _ => Set.empty[ObjectLevelVariable]
  }) ++ sub.freeVars

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = v match {
    case Const(_) => At(v, sub.groundBy(exceptions))
    case Var(_) if exceptions.contains(v) => At(v, sub.groundBy(exceptions))
    case Var(n) => At(Const(n), sub.groundBy(exceptions))
  }

  override def getConcepts: Set[ConceptConstant] = sub.getConcepts

  override def getRoles: Set[NamedRole] = sub.getRoles

  override def getIndividuals: Set[String] = (v match {
    case Const(n) => Set(n)
    case _ => Set.empty
  }) ++ sub.getIndividuals
}
