package src.ship

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 13.06.2012
 * Time: 17:11
 */

trait ObjectLevelVariable {
  def freeVars: Set[ObjectLevelVariable] = Set.empty
}

trait GroundableObject[+T] {
  def groundBy(exceptions: Set[ObjectLevelVariable]): T

  def ground: T = groundBy(Set.empty[ObjectLevelVariable])
}

sealed abstract class FOLTL[+TYPE <: ObjectLevelVariable with GroundableObject[TYPE], +ATOM <: ObjectLevelVariable with GroundableObject[ATOM]] extends ObjectLevelVariable with GroundableObject[FOLTL[TYPE, ATOM]] {
  def isOr: Option[(FOLTL[TYPE, ATOM], FOLTL[TYPE, ATOM])] = None
}

case object True extends FOLTL[Nothing, Nothing] {
  override def toString = "true"

  override def freeVars: Set[ObjectLevelVariable] = Set.empty[ObjectLevelVariable]

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = this
}

case class Atom[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](atom: ATOM) extends FOLTL[TYPE, ATOM] {
  override def toString = atom.toString

  override def freeVars: Set[ObjectLevelVariable] = atom.freeVars

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = Atom(atom.groundBy(exceptions))
}

case class RunProg[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](prog: DLProgram) extends FOLTL[TYPE, ATOM] {
  override def toString = "run " + prog

  override def freeVars = prog.freeVars

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = RunProg(prog.groundBy(exceptions))
}

case class DefPred[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](pred: Pred) extends FOLTL[TYPE, ATOM] {
  override def toString = pred.toString

  override def freeVars = pred.freeVars

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = DefPred(pred.groundBy(exceptions))
}

case class Not[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](sub: FOLTL[TYPE, ATOM]) extends FOLTL[TYPE, ATOM] {
  override def toString = {
    this match {
      case Not(True) => "False"
      case Not(U(True, Not(phi))) => "G(" + phi + ")"
      case Not(All(v, c, Not(phi))) => "\u2203" + v + ":" + c + " . " + phi
      case Not(And(Not(Not(phi1)), Not(phi2))) => phi1 + " \u21D2 " + phi2
      case Not(And(Not(phi1), Not(phi2))) => phi1 + " \u2228 " + phi2
      case Not(And(Not(phi1), phi2)) => phi1 + " \u2228 " + Not(phi2)
      case Not(And(phi1, Not(phi2))) => "\u00AC(" + phi1 + ") \u2228 " + phi2
      case _ => "\u00AC(" + sub + ")"
    }
  }

  override def isOr: Option[(FOLTL[TYPE, ATOM], FOLTL[TYPE, ATOM])] = {
    this match {
      case Not(And(phi1, phi2)) => Some(Not(phi1), Not(phi2))
      case _ => None
    }
  }

  override def freeVars: Set[ObjectLevelVariable] = sub.freeVars

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = Not(sub.groundBy(exceptions))
}

case class And[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](sub1: FOLTL[TYPE, ATOM], sub2: FOLTL[TYPE, ATOM]) extends FOLTL[TYPE, ATOM] {
  override def toString = sub1 + " \u2227 " + sub2

  override def freeVars: Set[ObjectLevelVariable] = sub1.freeVars ++ sub2.freeVars

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = And(sub1.groundBy(exceptions), sub2.groundBy(exceptions))
}

case class All[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](variable: ObjectLevelVariable, vartype: TYPE, sub: FOLTL[TYPE, ATOM]) extends FOLTL[TYPE, ATOM] {
  override def toString = "\u2200 " + variable + ":" + vartype + " . " + sub

  override def freeVars: Set[ObjectLevelVariable] = (sub.freeVars - variable) ++ vartype.freeVars

  override def groundBy(exceptions: Set[ObjectLevelVariable]) =
    All(variable, vartype.groundBy(exceptions union Set[ObjectLevelVariable](variable)), sub.groundBy(exceptions union Set[ObjectLevelVariable](variable)))
}

case class X[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](sub: FOLTL[TYPE, ATOM]) extends FOLTL[TYPE, ATOM] {
  override def freeVars: Set[ObjectLevelVariable] = sub.freeVars

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = X(sub.groundBy(exceptions))
}

case class U[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](sub1: FOLTL[TYPE, ATOM], sub2: FOLTL[TYPE, ATOM]) extends FOLTL[TYPE, ATOM] {
  override def toString = {
    this match {
      case U(True, phi) => "F(" + phi + ")"
      case U(phi1, phi2) => phi1 + " U " + phi2
    }
  }

  override def freeVars: Set[ObjectLevelVariable] = sub1.freeVars ++ sub2.freeVars

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = U(sub1.groundBy(exceptions), sub2.groundBy(exceptions))
}

trait FOLTLParser extends ProgramAndMonitorParser {}

object FOLTL extends FOLTLParser {
  def Or[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](f: FOLTL[TYPE, ATOM], g: FOLTL[TYPE, ATOM]): FOLTL[TYPE, ATOM] = {
    Not(And(Not(f), Not(g)))
  }


  def Impl[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](f: FOLTL[TYPE, ATOM], g: FOLTL[TYPE, ATOM]): FOLTL[TYPE, ATOM] = {
    Or(Not(f), g)
  }

  def Ex[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](variable: ObjectLevelVariable, vartype: TYPE, sub: FOLTL[TYPE, ATOM]) = {
    Not(All(variable, vartype, Not(sub)))
  }

  def F[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](formula: FOLTL[TYPE, ATOM]): FOLTL[TYPE, ATOM] = {
    U(True, formula)
  }

  def G[TYPE <: ObjectLevelVariable with GroundableObject[TYPE], ATOM <: ObjectLevelVariable with GroundableObject[ATOM]](formula: FOLTL[TYPE, ATOM]): FOLTL[TYPE, ATOM] = {
    Not(F[TYPE, ATOM](Not(formula)))
  }

  def False = Not(True)

  def queryOntology(o: Ontology, formula: FOLTL[Concept, ABoxFormula]): List[(FOLTL[Concept, ABoxFormula], Substitution)] = {
    formula match {
      case Atom(phi) => phi match {
        case DifferentFromAssertion(Const(n), Const(m)) => List((if (n != m) True else False, Id()))
        case SameAsAssertion(Const(n), Const(m)) => List((if (n == m) True else False, Id()))
        case _ => o.query(phi) match {
          case Nil => List((False, Id()))
          case l => l map (s => (True, s))
        }
      }
      case Not(Atom(phi)) => phi match {
        case DifferentFromAssertion(u, v) => (u, v) match {
          case (Const(n), Const(m)) => List((if (n == m) True else False, Id()))
          case _ => queryOntology(o, Atom[Concept, ABoxFormula](SameAsAssertion(u, v)))
        }
        case SameAsAssertion(u, v) => (u, v) match {
          case (Const(n), Const(m)) => List((if (n != m) True else False, Id()))
          case _ => queryOntology(o, Atom[Concept, ABoxFormula](DifferentFromAssertion(u, v)))
        }
        case ConceptAssertion(n, c) => n match {
          case v@Var(name) =>
            val allind = o.allIndividuals filterNot o.allIndividuals(c).contains
            //System.out.println("Allind = "+allind)
            val res = allind match {
              case _ :: _ => allind map (iname => {
                val subst = /(Const(iname), v)
                (True, subst)
              })
              case Nil => List((False, Id()))
            }
            res
          case _ => List((if (o.isUnsatisfiable(phi)) True else False, Id()))
        }
        case RoleAssertion(n, m, r) =>
          val rsignature = o.getRoleSignature(r)
          val substsn: List[Substitution] = n match {
            case v@Var(name) => o.allIndividuals(rsignature._1) map (iname => /(Const(iname), v))
            case _ => List(Id())
          }
          val substsm: List[Substitution] = m match {
            case v@Var(name) => o.allIndividuals(rsignature._2) map (iname => /(Const(iname), v))
            case _ => List(Id())
          }
          val res = substsn flatMap ((nsubst: Substitution) => substsm flatMap ((msubst: Substitution) => {
            val combsubst = CompSubst(nsubst, msubst)
            if (o.isUnsatisfiable(combsubst.apply(phi))) List((True, combsubst))
            else List.empty
          }))
          if (res.isEmpty)
            List((False, Id()))
          else res
        case _ => throw new Exception("queryOntology2: " + formula)
      }
      case _ => throw new Exception("queryOntology: " + formula)
    }
  }

  def individualsOntology(o: Ontology, c: Concept): List[String] =
    o.allIndividuals(c)

  type Result = (FOLTL[Concept, ABoxFormula], Substitution)

  def progress(o: Ontology, formula: FOLTL[Concept, ABoxFormula], decls: Map[String, DLDeclaration]): FOLTL[Concept, ABoxFormula] = {

    progress_intern(o, formula, decls) match {
      case (phi, subst) :: xs => phi
      case Nil => formula
    }
  }

  def progress_intern(o: Ontology, formula: FOLTL[Concept, ABoxFormula], decls: Map[String, DLDeclaration]): List[Result] = {
    val res = {
      formula match {
        case True => List((True, Id()))
        case Not(True) => List((False, Id()))
        case Atom(a) => queryOntology(o, Atom(a))
        case Not(Atom(a)) => queryOntology(o, Not(Atom(a)))
        case Not(i@DefPred(Pred(n, actpars))) => progress_intern_defpred(o, i, decls, negated = true)
        case RunProg(_) => List((formula, Id()))
        case i@DefPred(Pred(n, actpars)) => progress_intern_defpred(o, i, decls, negated = false)
        case And(a, b) =>
          (progress_intern(o, a, decls)
            flatMap {
            case (ae, substa) =>
              (progress_intern(o, substa.applyF(b), decls)
                map {
                case (be, substb) => (And(ae, be), CompSubst(substb, substa))
              })
          })
        case U(a, b) => progress_intern(o, a, decls) flatMap {
          case (ae, substa) =>
            (progress_intern(o, substa.applyF(b), decls)
              map {
              case (be, substb) => (Or(be, And(ae, U(a, b))), CompSubst(substb, substa))
            })
        }
        case Not(And(a, b)) =>
          (progress_intern(o, Not(a), decls)
            flatMap {
            case (ae, substa) =>
              (progress_intern(o, substa.applyF(Not(b)), decls)
                map {
                case (be, substb) => (Or(ae, be), CompSubst(substb, substa))
              })
          })
        case All(v: Var, c: Concept, f: FOLTL[Concept, ABoxFormula]) =>
          progress_intern(o, (individualsOntology(o, c) :\ (True: FOLTL[Concept, ABoxFormula]))
            ((n, phi) => And[Concept, ABoxFormula](/(Const(n), v).applyF(f), phi)), decls)
        case Not(All(v: Var, c: Concept, f: FOLTL[Concept, ABoxFormula])) =>
          val allindc = o.allIndividuals(c)
          allindc match {
            case Nil => List((False, Id()))
            case x :: xs =>
              val negf: FOLTL[Concept, ABoxFormula] = f match {
                case Not(g) => g
                case _ => Not(f)
              }
              val disjunction: FOLTL[Concept, ABoxFormula] = (xs :\ /(Const(x), v).applyF(negf))((x, psi) => Or(/(Const(x), v).applyF(negf), psi))
              progress_intern(o, disjunction, decls)
          }
        case Not(Not(phi)) => progress_intern(o, phi, decls)
        /* not(phi U psi) = not psi /\ (not phi \/ X(not(phi U psi))) */
        case Not(U(phi1, phi2)) => progress_intern(o, And(Not(phi2), Or(Not(phi1), X(Not(U(phi1, phi2))))), decls)
        case X(f: FOLTL[Concept, ABoxFormula]) => List((f, Id()))
        case _ => throw new Exception("no progress for " + formula)
      }
    }
    //System.out.println(""+formula+" => "+res)
    res map (p => (simplify(p._1), p._2))
  }


  def progress_intern_defpred(o: Ontology, formula: DefPred[Concept, ABoxFormula],
                              decls: Map[String, DLDeclaration], negated: Boolean): List[Result] = {
    formula match {
      case DefPred(Pred(n, actpars)) =>
        if (decls contains n) {
          decls get n match {
            case Some(Predicate(name, fpars, desc, b)) =>
              Pred(n, actpars).expand(Predicate(name, fpars, desc, b)) match {
                case MonitorExec(phi, _, psi) =>
                  //println("Progress1 negated = "+negated+";"+(if (negated) Not(psi) else psi).toString)
                  val res = progress_intern(o, if (negated) Not(psi) else psi, decls)
                  //println("Progress2 "+formula+" "+res)
                  res
                case _ => List((formula, Id()))
              }
            case _ => List((formula, Id()))
          }
        }
        else List((formula, Id()))
    }
  }


  def diagnose_formula(o: Ontology, phi: FOLTL[Concept, ABoxFormula]): FOLTL[Concept, ABoxFormula] = {
    phi match {
      case True => True
      case Not(Not(psi)) => Not(diagnose_formula(o, psi))
      case Not(And(phi1, phi2)) => Not(And(diagnose_formula(o, Not(phi1)), diagnose_formula(o, Not(phi2))))
      case And(phi1, phi2) => And(diagnose_formula(o, phi1),
        diagnose_formula(o, phi2))
      case All(v: Var, c: Concept, f: FOLTL[Concept, ABoxFormula]) =>
        diagnose_formula(o, (individualsOntology(o, c) :\ (True: FOLTL[Concept, ABoxFormula]))
          ((n, phi) => And[Concept, ABoxFormula](/(Const(n), v).applyF(f), phi)))
      case Not(All(v: Var, c: Concept, f: FOLTL[Concept, ABoxFormula])) =>
        diagnose_formula(o, Not(f))
      case U(a, b) => Or(diagnose_formula(o, a), diagnose_formula(o, b))
      case Not(U(phi1, phi2)) => And(diagnose_formula(o, Not(phi2)), diagnose_formula(o, Not(phi1)))
      case X(f) => f
      case f => f
    }
  }

  /*
  type PairFormulas = Pair[List[FOLTL[Concepprogress_internt, ABoxFormula]],List[FOLTL[Concept, ABoxFormula]]]


  def Merge(p1:PairFormulas,p2:PairFormulas) =  (p1._1 ++ p2._1,p1._2 ++ p2._2)

  def progress_parallel(o1: Ontology, o2:Ontology, phi: FOLTL[Concept, ABoxFormula])
  :Pair[List[FOLTL[Concept, ABoxFormula]],List[FOLTL[Concept, ABoxFormula]]] = {

    if (progress_intern(o1,phi)!=False && progress_intern(o2,phi)==False) {
      phi match {
        case And(phi1,phi2) => Merge(progress_parallel(o1,o2,ph1),progress_parallel(o1,o2,ph2))
        case Not(And(phi1,phi2)) => Merge(progress_parallel(o1,o2,Not(ph1)),progress_parallel(o1,o2,Not(ph2)))
        case U(phi1,phi2) => Merge(progress_parallel(o1,o2,ph1),progress_parallel(o1,o2,ph2))
        case Not(Not(phi)) => progress_parallel(o1,o2,phi)
        case All(v,c,phi) => {
          val o1c = individualsOntology(o1,c)
          val o2c = individualsOntology(o2,c)
          val common1 = ((o1c filter (o1 => (o2c exists (o2 => (o1==o2))))) map (o => progress_parallel(o1,o2,/(o,v)(phi))))
          val common = (common1 :\ (Nil,Nil)) Merge;
          progress_intern(o1,phi)


        }
      }
  }
    (Nil,Nil)
  }
  */

  def isNegationOf(phi1: FOLTL[Concept, ABoxFormula], phi2: FOLTL[Concept, ABoxFormula]): Boolean = {
    (phi1 == Not(phi2)) || (Not(phi1) == phi2)
  }

  def simplify(phi: FOLTL[Concept, ABoxFormula]): FOLTL[Concept, ABoxFormula] = {
    phi match {
      case And(p1, p2) =>
        val phi1 = simplify(p1)
        val phi2 = simplify(p2)
        if (phi1 == False || phi2 == False) False
        else if (phi1 == True) phi2
        else if (phi2 == True) phi1
        else if (phi1 == phi2) phi1
        else if (isNegationOf(phi1, phi2)) False
        else phi2 match {
          case And(phi3, phi4) =>
            if (phi1 == phi3) And(phi1, phi4)
            else if (phi1 == phi4) And(phi1, phi3)
            else And(phi1, phi2)
          case _ => And(phi1, phi2)
        }
      case Not(p1) =>
        val phi1 = simplify(p1)
        if (phi1 == True) False
        else if (phi1 == False) True
        else phi1 match {
          case Not(phi11) => phi11
          case _ => Not(phi1)
        }
      case All(v, c, p1) =>
        val phi1 = simplify(p1)
        if (phi1 == True) True
        else if (phi1 == False) False
        else All(v, c, phi1)

      case X(p1) =>
        val phi1 = simplify(p1)
        if (phi1 == True || phi1 == False) phi1
        else X(phi1)

      case U(p1, p2) =>
        val phi1 = simplify(p1)
        val phi2 = simplify(p2)
        if (phi2 == False) False
        else if (phi2 == True) phi2
        else if (phi1 == False) phi1
        else U(phi1, phi2)
      case _ => phi
    }
  }
}



