package src.ship

import org.semanticweb.owlapi.model._
import org.semanticweb.owlapi.vocab.PrefixOWLOntologyFormat
import scala.collection.{mutable, JavaConversions}

/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 7/18/12
 * Time: 3:46 PM
 * converts an ontology in the internal format to an owl ontology that can be used e.g. in combination with pellet
 */

object Ontology2OWL extends JavaLogging {

  /**
   * converts an ontology in the internal format to an owl ontology
   */
  def convert(ont: Ontology, pm: PrefixManager, manager: OWLOntologyManager): OWLOntology = {

    val ontology = if (ont.isBooleanAbox) ont.toClassicAbox else ont
    val owlont: OWLOntology = manager.createOntology(IRI.create(pm.getDefaultPrefix + "ship"))
    val form = new PrefixOWLOntologyFormat()
    form.setDefaultPrefix(pm.getDefaultPrefix)
    manager.setOntologyFormat(owlont, form)

    for (tbox: TBoxFormula <- ontology.tbox) tbox2OWL(tbox, owlont, pm, manager)
    //val classes:Set[OWLClass] = JavaConversions.asScalaSet(owlont.getClassesInSignature)

    //println("classes are "+classes+" "+owlont.getClassesInSignature)
    //addComplementClasses(owlont, classes.toList, manager, manager.getOWLDataFactory)
    //addDistinctnessAxioms(owlont, computeTopClasses(owlont), manager)

    for (rbox: RBoxFormula <- ontology.rbox) rbox2OWL(rbox, owlont, pm, manager)

    for (abox: ABoxFormula <- ontology.abox) {
      // println("Converting " + abox)
      abox2OWL(abox, owlont, pm, manager)
    }

    //closeWorld(ontology.abox, owlont, pm, manager)

    //@todo maybe this was not a good idea?
    //addComplementClasses(owlont, JavaConversions.asScalaSet(owlont.getIndividualsInSignature),manager, manager.getOWLDataFactory)


    for((k,v) <- ont.annotationMap) k match {
      case x : OWLOntology => manager.applyChange(new AddOntologyAnnotation(owlont, v))
      case x : OWLAxiom => manager.applyChange(new RemoveAxiom(owlont, x.getAxiomWithoutAnnotations))
        manager.applyChange(new AddAxiom(owlont, x))
      case _ =>
    }
    owlont
  }

  /**
   * closes the world by declaring thing to be equivalent to all individuals in the list
   * and that all individuals denote different things (unique name assumption)
   */
  def closeWorld(aBoxFormulas: List[ABoxFormula], ontology: OWLOntology, pm: PrefixManager, manager: OWLOntologyManager) {
    val factory = manager.getOWLDataFactory
    val individuals = findIndividuals(aBoxFormulas, pm, factory)
    //println("Individuals in ontology: "+individuals)

    if (!individuals.isEmpty) {
      manager.addAxiom(ontology, factory.getOWLDifferentIndividualsAxiom(JavaConversions.mutableSetAsJavaSet(individuals)))
      manager.addAxiom(ontology, factory.getOWLEquivalentClassesAxiom(factory.getOWLThing,
        factory.getOWLObjectOneOf(JavaConversions.mutableSetAsJavaSet(individuals))))
    }
  }

  /**
   * finds from all aboxformulas those that are individual assertions and collects them in a set
   */
  def findIndividuals(aBoxFormulas: List[ABoxFormula], pm: PrefixManager, factory: OWLDataFactory): mutable.Set[OWLIndividual] = {
    val res = scala.collection.mutable.Set[OWLIndividual]()
    for (a <- aBoxFormulas) {
      a match {
        case ConceptAssertion(Const(n), c) => res.add(factory.getOWLNamedIndividual(":" + n, pm))
        case RoleAssertion(s, t, _) =>
          //println(""+s+s.isInstanceOf[Const]+t+t.isInstanceOf[Const])
          s match {
            case Const(n) => res.add(factory.getOWLNamedIndividual(":" + n, pm))
            case _ =>
          }
          t match {
            case Const(n) => res.add(factory.getOWLNamedIndividual(":" + n, pm))
            case _ =>
          }
        //println(res)
        case DifferentFromAssertion(s, t) =>
          s match {
            case Const(n) => res.add(factory.getOWLNamedIndividual(":" + n, pm))
            case _ =>
          }
          t match {
            case Const(n) => res.add(factory.getOWLNamedIndividual(":" + n, pm))
            case _ =>
          }
        case SameAsAssertion(s, t) =>
          s match {
            case Const(n) => res.add(factory.getOWLNamedIndividual(":" + n, pm))
            case _ =>
          }
          t match {
            case Const(n) => res.add(factory.getOWLNamedIndividual(":" + n, pm))
            case _ =>
          }
        case _ => throw new Exception("findIndividuals: " + a)
      }
    }
    res
  }

  /**
   * adds the specified rboxformula to the specified owl ontology
   * @param rBoxFormula: a rbox formula
   */
  def rbox2OWL(rBoxFormula: RBoxFormula, ontology: OWLOntology, pm: PrefixManager, manager: OWLOntologyManager) {
    val factory = manager.getOWLDataFactory
    rBoxFormula match {
      case RoleEquality(InverseRole(x) :: List(), role) =>
        manager.addAxiom(
          ontology,
          factory.getOWLInverseObjectPropertiesAxiom(
            role2OWLPropExpr(role, pm, factory), role2OWLPropExpr(x, pm, factory)
          ))
      case RoleEquality(rolecomposition, role) =>
        throw new Exception("translation of rbox equality not supported: " + rBoxFormula)
      case RoleInclusion(rolecomposition, role) =>
        rolecomposition match {
          case x :: List() => manager.addAxiom(ontology,
            factory.getOWLSubObjectPropertyOfAxiom(role2OWLPropExpr(x, pm, factory),
              role2OWLPropExpr(role, pm, factory)))
          case _ => manager.addAxiom(ontology, factory.getOWLSubPropertyChainOfAxiom(
            JavaConversions.seqAsJavaList(rolecomposition map ((x: Role) => role2OWLPropExpr(x, pm, factory))),
            role2OWLPropExpr(role, pm, factory)))
        }
      case RoleSym(role) => manager.addAxiom(ontology,
        factory.getOWLSymmetricObjectPropertyAxiom(role2OWLPropExpr(role, pm, factory)))
      case RoleTra(role) => manager.addAxiom(ontology,
        factory.getOWLTransitiveObjectPropertyAxiom(role2OWLPropExpr(role, pm, factory)))
      case RoleRef(role) => manager.addAxiom(ontology,
        factory.getOWLReflexiveObjectPropertyAxiom(role2OWLPropExpr(role, pm, factory)))
      case RoleIrr(role) => manager.addAxiom(ontology,
        factory.getOWLIrreflexiveObjectPropertyAxiom(role2OWLPropExpr(role, pm, factory)))
      case RoleAsym(role) => manager.addAxiom(ontology,
        factory.getOWLAsymmetricObjectPropertyAxiom(role2OWLPropExpr(role, pm, factory)))
      case RoleFun(role@DataRole(dr)) =>
        val roleProp: OWLDataProperty = factory.getOWLDataProperty(":" + dr, pm)
        manager.addAxiom(ontology, factory.getOWLFunctionalDataPropertyAxiom(roleProp))
      case RoleFun(role) => manager.addAxiom(ontology,
        factory.getOWLFunctionalObjectPropertyAxiom(role2OWLPropExpr(role, pm, factory)))
      case RoleInvFun(role) => manager.addAxiom(ontology,
        factory.getOWLInverseFunctionalObjectPropertyAxiom(role2OWLPropExpr(role, pm, factory)))
      case RoleDis(_, _) => throw new Exception("translation of disjoint roles not supported: " + rBoxFormula)
      case RoleDeclaration(role, conceptDomain, conceptRange) => {
        role match {
          case DataRole(n) =>
            val roleProp: OWLDataProperty = factory.getOWLDataProperty(":" + n, pm)

            conceptDomain match {
              case UniversalConcept => {}
              case _ =>
                val domain = concept2PlainOWLClassExpr(conceptDomain, pm, factory)
                manager.addAxiom(ontology, factory.getOWLDataPropertyDomainAxiom(roleProp, domain))
            }
            conceptRange match {
              case UniversalConcept => {}
              case _ =>
                val range = concept2OWLDataRange(conceptRange, pm, factory)
                manager.addAxiom(ontology, factory.getOWLDataPropertyRangeAxiom(roleProp, range))
            }
          case NamedRole(n) =>
            val roleprop: OWLObjectProperty = factory.getOWLObjectProperty(":" + n, pm)
            val domain = concept2PlainOWLClassExpr(conceptDomain, pm, factory)
            val range = concept2PlainOWLClassExpr(conceptRange, pm, factory)
            manager.addAxiom(ontology, factory.getOWLObjectPropertyDomainAxiom(roleprop, domain))
            manager.addAxiom(ontology, factory.getOWLObjectPropertyRangeAxiom(roleprop, range))
          case _ => throw new Exception("Cannot translated non-constant role declaration " + role)
        }
      }
      case _ => throw new Exception("translation of concept not supported (yet)?" + rBoxFormula + " " + rBoxFormula.getClass)
    }
  }

  /**
   * adds the specified tboxformula to the specified owl ontology
   * @param tbox: a tbox formula
   */
  def tbox2OWL(tbox: TBoxFormula, ontology: OWLOntology, pm: PrefixManager, manager: OWLOntologyManager) {
    val factory = manager.getOWLDataFactory
    tbox match {
      case ConceptEquality(ConceptConstant(lhs), rhs) =>
        manager.addAxiom(ontology, factory.getOWLEquivalentClassesAxiom(
          factory.getOWLClass(lhs, pm), concept2PlainOWLClassExpr(rhs, pm, factory)))
      case ConceptEquality(NotConcept(UniversalConcept), AndConcept(c1, c2)) =>
        manager.addAxiom(ontology, factory.getOWLDisjointClassesAxiom(
          concept2PlainOWLClassExpr(c1, pm, factory),
          concept2PlainOWLClassExpr(c2, pm, factory)))
      case ConceptInclusion(ConceptConstant(sub), sup) =>
        manager.addAxiom(ontology, factory.getOWLSubClassOfAxiom(
          factory.getOWLClass(sub, pm), concept2PlainOWLClassExpr(sup, pm, factory)))
      case _ =>
        throw new Exception("Cannot add TBOX " + tbox + ". Is LHS atomic?")
    }
  }

  /**
   * adds the specified aboxformula to the specified owl ontology
   * @param abox: a abox formula
   */
  def abox2OWL(abox: ABoxFormula, ontology: OWLOntology, pm: PrefixManager, manager: OWLOntologyManager) = try
    manager.addAxiom(ontology, aboxformula2OWL(abox, ontology, pm, manager))
  catch {
    case e: Exception => log.error("converting abox " + abox + " failed: " + e.toString)
  }

  def aboxformula2OWL(phi: ABoxFormula, ontology: OWLOntology, pm: PrefixManager, manager: OWLOntologyManager): OWLAxiom = {
    val factory = manager.getOWLDataFactory
    phi match {
      case ConceptAssertion(Var(name), concept) =>
        throw new Exception("translation of concept not supported (yet)?" + concept)
      case ConceptAssertion(Const(name), concept) =>
        val ind = factory.getOWLNamedIndividual(":" + name, pm)
        factory.getOWLClassAssertionAxiom(concept2PlainOWLClassExpr(concept, pm, factory), ind)
      case RoleAssertion(Const(src), Literal(trg), role) => role match {
        case DataRole(n) =>
          val srcind = factory.getOWLNamedIndividual(":" + src, pm)
          val dp = factory.getOWLDataProperty(":" + n, pm)
          val rset = JavaConversions.asScalaSet(ontology.getDataPropertyRangeAxioms(dp))
          log.debug("Rset for " + dp + " = " + rset)
          if (rset.size == 1) {
            val dt = rset.head.getRange.asOWLDatatype().getBuiltInDatatype
            val strippedLit = if (trg.length > 1 && trg.startsWith("\"") && trg.endsWith("\""))
              trg.tail.init
            else trg
            if (dt.isInLexicalSpace(strippedLit)) {

              val lit = factory.getOWLLiteral(strippedLit, dt)
              val res = factory.getOWLDataPropertyAssertionAxiom(dp, srcind, lit)
              log.debug(n + " --> " + res)
              res
            } else throw new Exception("illegal literal of type " + dt + ": " + strippedLit)
          } else throw new Exception("unknown data property: " + n)
        case _ => throw new Exception("cannot process complex role: " + role)
      }
      case RoleAssertion(Const(src), Const(trg), role) => {
        val srcind = factory.getOWLNamedIndividual(":" + src, pm)
        val trgind = factory.getOWLNamedIndividual(":" + trg, pm)

        role match {
          case NamedRole("SameAs") => factory.getOWLSameIndividualAxiom(srcind, trgind)
          case NamedRole("Different") => factory.getOWLDifferentIndividualsAxiom(srcind, trgind)
          case NotRole(r) =>
            factory.getOWLNegativeObjectPropertyAssertionAxiom(
              role2OWLPropExpr(r, pm, factory), srcind, trgind
            )
          case DataRole(n) =>
            val dp = factory.getOWLDataProperty(":" + n, pm)
            val rset = JavaConversions.asScalaSet(ontology.getDataPropertyRangeAxioms(dp))
            log.debug("Rset for " + dp + " = " + rset)
            if (rset.size == 1) {
              val dt = rset.head.getRange.asOWLDatatype().getBuiltInDatatype
              val strippedLit = if (trg.length > 1 && trg.startsWith("\"") && trg.endsWith("\""))
                trg.tail.init
              else trg
              if (dt.isInLexicalSpace(strippedLit)) {

                val lit = factory.getOWLLiteral(strippedLit, dt)
                val res = factory.getOWLDataPropertyAssertionAxiom(datarole2OWLPropExpr(role, pm, factory), srcind, lit)
                log.debug(n + " --> " + res)
                res
              } else throw new Exception("illegal literal of type " + dt + ": " + strippedLit)
            } else throw new Exception("unknown data property: " + n)


          case _ => factory.getOWLObjectPropertyAssertionAxiom(role2OWLPropExpr(role, pm, factory), srcind, trgind)
        }
      }
      case DifferentFromAssertion(Const(o1), Const(o2)) =>
        val i1 = factory.getOWLNamedIndividual(":" + o1, pm)
        val i2 = factory.getOWLNamedIndividual(":" + o2, pm)
        factory.getOWLDifferentIndividualsAxiom(i1, i2)
      case SameAsAssertion(Const(o1), Const(o2)) =>
        val i1 = factory.getOWLNamedIndividual(":" + o1, pm)
        val i2 = factory.getOWLNamedIndividual(":" + o2, pm)
        factory.getOWLSameIndividualAxiom(i1, i2)
      case _ => throw new Exception("cannot process " + phi)
    }
  }

  /**
   * computes all classes in the signature that do not have predecessor classes
   * @param ont: an owl ontology
   */
  def computeTopClasses(ont: OWLOntology): List[OWLClass] = {
    val iterator = ont.getClassesInSignature.iterator()
    var topclasses: List[OWLClass] = List.empty[OWLClass]
    while (iterator.hasNext) {
      val c: OWLClass = iterator.next()
      if (!c.isDefined(ont) && c.getEquivalentClasses(ont).isEmpty && c.getSuperClasses(ont).isEmpty) topclasses = topclasses ++ List(c)
    }
    topclasses
  }


  /**
   * adds for all individuals distinctness axioms and declares the specified classes to be distinct
   * thing is declared to be the disjoint union of the base classes
   * @param ont: an owl ontology
   * @param baseclasses: a list of classes that are defined to be distinct
   */
  def addDistinctnessAxioms(ont: OWLOntology, baseclasses: List[OWLClass], manager: OWLOntologyManager) {
    val factory = manager.getOWLDataFactory
    manager.addAxiom(ont, factory.getOWLDifferentIndividualsAxiom(ont.getIndividualsInSignature))
    /*manager.addAxiom(ont, factory.getOWLDisjointClassesAxiom(JavaConversions.setAsJavaSet(baseclasses.toSet[OWLClass])))
     manager.addAxiom(ont, factory.getOWLDisjointUnionAxiom(factory.getOWLClass(IRI.create("owl:Thing")),
      JavaConversions.setAsJavaSet(baseclasses.toSet[OWLClass])))
      */
  }

  /**
   * adds complement classes for all classes in the list (except of owl:Thing). The complement class name has the
   * prefix "Neg"
   * @param ont: the owl ontology to which the negations are to be added
   * @param classes: the classes for which a complement is to be added
   */
  def addComplementClasses(ont: OWLOntology, classes: List[OWLClass], manager: OWLOntologyManager, factory: OWLDataFactory) {
    log.info("add complement classes")
    val it = classes.iterator
    while (it.hasNext) {
      val c: OWLClass = it.next()
      if (!c.isOWLThing)
        manager.addAxiom(ont, factory.getOWLEquivalentClassesAxiom(
          internalComplementClass(c, factory),
          factory.getOWLObjectComplementOf(c)))
    }
  }

  /**
   * adds complement classes for all named individuals in the list (except of owl:Thing). The complement class name has the
   * prefix "Neg"
   * @param ont: the owl ontology to which the negations are to be added
   * @param classes: the classes for which a complement is to be added
   */
  def addComplementClasses(ont: OWLOntology, classes: mutable.Set[OWLNamedIndividual], manager: OWLOntologyManager, factory: OWLDataFactory) {
    log.info("add complement classes for individuals")
    val it = classes.iterator
    while (it.hasNext) {
      val c: OWLNamedIndividual = it.next()
      manager.addAxiom(ont, factory.getOWLEquivalentClassesAxiom(
        internalComplementClass(c, factory),
        factory.getOWLObjectComplementOf(factory.getOWLObjectOneOf(c))))
    }
  }

  /**
   * computes the internal class complement of a given class (that was added earlier of the translation process)
   */
  def internalComplementClass(c: OWLLogicalEntity, factory: OWLDataFactory): OWLClass = {
    //println("internal complement for "+c)
    factory.getOWLClass(IRI.create(c.getIRI.getNamespace + "Neg" + c.getIRI.getFragment))
  }

  /**
   * converts a role to a OWLPropertyExpression
   */
  def role2OWLPropExpr(role: Role, pm: PrefixManager, factory: OWLDataFactory):
  OWLObjectPropertyExpression = {
    role match {
      case NamedRole(n) => factory.getOWLObjectProperty(n, pm)
      case InverseRole(r) => factory.getOWLObjectInverseOf(role2OWLPropExpr(r, pm, factory))
      case _ => log.warning("translation of role to property not supported: " + role)
        factory.getOWLTopObjectProperty
    }
  }

  def datarole2OWLPropExpr(role: Role, pm: PrefixManager, factory: OWLDataFactory):
  OWLDataPropertyExpression = {
    role match {
      case DataRole(n) => factory.getOWLDataProperty(n, pm)
      case _ => throw new Exception("translation of data role to property not supported: " + role)
    }
  }

  def concept2PlainOWLClassExpr(c: Concept, pm: PrefixManager, f: OWLDataFactory): OWLClassExpression = {
    concept2OWLClassExpr(c, pm, f, usenegabbrev = false)
  }

  def concept2OWLDataRange(c: Concept, pm: PrefixManager, f: OWLDataFactory): OWLDataRange = {
    c match {
      case DataConcept(name, _, _) => f.getOWLDatatype("xsd:" + name, pm)
      case _ => throw new Exception("translation of concept to data range not supported: " + c)
    }
  }

  /**
   * converts a concept to a OWLClassExpression
   * @param concept: concept to be translated
   * @param pm: prefix manager that adds the prefix for the ontology
   * @param factory: factory to be used to construct the class expressions
   * @param usenegabbrev: if true, complement classes are replaced by the corresponding defined complement class that
   *                    was added during the translation process
   */
  def concept2OWLClassExpr(concept: Concept, pm: PrefixManager, factory: OWLDataFactory, usenegabbrev: Boolean): OWLClassExpression =
    concept match {
      case UniversalConcept => factory.getOWLThing
      case NotConcept(UniversalConcept) => factory.getOWLNothing
      case ConceptConstant(name) => {
        val s = Builtin.upCase1(name)
        factory.getOWLClass(s, pm)
      }
      case ConceptVariable(s) => throw new Exception("translation of concept not supported (yet)?")
      case NotConcept(c) =>
        if (usenegabbrev) {
          //println("Concept2OWL with use=true")

          c match {
            case ConceptConstant(s) => internalComplementClass(factory.getOWLClass(s, pm), factory)
            case Nominal(Const(n)) => internalComplementClass(factory.getOWLNamedIndividual(":" + n, pm), factory)
            case _ => throw new Exception("internal concepts only defined for atomic concepts? " + c)
          }
        }
        else factory.getOWLObjectComplementOf(concept2OWLClassExpr(c, pm, factory, usenegabbrev))
      case AndConcept(c1, c2) => factory.getOWLObjectIntersectionOf(concept2OWLClassExpr(c1, pm, factory, usenegabbrev),
        concept2OWLClassExpr(c2, pm, factory, usenegabbrev))
      case OrConcept(c1, c2) => factory.getOWLObjectUnionOf(concept2OWLClassExpr(c1, pm, factory, usenegabbrev),
        concept2OWLClassExpr(c2, pm, factory, usenegabbrev))
      case ForallRole(r, c) =>
        r match {
          case DataRole(n) =>
            val role = datarole2OWLPropExpr(r, pm, factory)
            factory.getOWLDataAllValuesFrom(role, concept2OWLDataRange(c, pm, factory))
          case _ =>
            val role = role2OWLPropExpr(r, pm, factory)
            factory.getOWLObjectAllValuesFrom(role, concept2OWLClassExpr(c, pm, factory, usenegabbrev))
        }
      case ExistsRole(r, c) =>
        r match {
          case DataRole(n) =>
            val role = datarole2OWLPropExpr(r, pm, factory)
            factory.getOWLDataSomeValuesFrom(role, concept2OWLDataRange(c, pm, factory))
          case _ =>
            val role = role2OWLPropExpr(r, pm, factory)
            factory.getOWLObjectSomeValuesFrom(role, concept2OWLClassExpr(c, pm, factory, usenegabbrev))
        }
      case Nominal(v) => v match {
        case _: Literal => throw new Exception("translation of nominal concepts with literal not supported: " + concept)
        case _ => factory.getOWLObjectOneOf(factory.getOWLNamedIndividual(":" + v.name, pm))
      }
      case MaxRestriction(r@DataRole(_), n, c@DataConcept(_, _, _)) =>
        factory.getOWLDataMaxCardinality(n, datarole2OWLPropExpr(r, pm, factory),
          concept2OWLDataRange(c, pm, factory))
      case MinRestriction(r@DataRole(_), n, c@DataConcept(_, _, _)) =>
        factory.getOWLDataMinCardinality(n, datarole2OWLPropExpr(r, pm, factory),
          concept2OWLDataRange(c, pm, factory))
      case MaxRestriction(r, n, c) =>
        factory.getOWLObjectMaxCardinality(n, role2OWLPropExpr(r, pm, factory),
          concept2OWLClassExpr(c, pm, factory, usenegabbrev))
      case MinRestriction(r, n, c) => factory.getOWLObjectMinCardinality(n, role2OWLPropExpr(r, pm, factory),
        concept2OWLClassExpr(c, pm, factory, usenegabbrev))
      case _ => log.warning("translation of concept to OWLClassExpr not supported: " + concept + concept.getClass)
        factory.getOWLThing
    }

  /**
   * converts a concept to a OWLClassExpression
   */
  def concept2SparqlString(concept: Concept, pm: PrefixManager, factory: OWLDataFactory): String = {
    concept match {
      case ConceptConstant(s) =>
        factory.getOWLClass(s, pm).toString
      case ConceptVariable(s) => "Class(?" + s + ")"
      case NotConcept(ConceptConstant(s)) =>
        log.info("Concept2Sparql with use=true")
        internalComplementClass(factory.getOWLClass(s, pm), factory).toString
        factory.getOWLObjectComplementOf(concept2PlainOWLClassExpr(ConceptConstant(s), pm, factory)).toString
      //factory.getOWLObjectComplementOf(factory.getOWLClass(s, pm)).toString
      case UniversalConcept => factory.getOWLThing.toString
      case NotConcept(UniversalConcept) => factory.getOWLNothing.toString
      case AndConcept(c1, c2) =>
        throw new Exception("translation of concept not supported (yet)?")
      case OrConcept(c1, c2) =>
        throw new Exception("translation of concept not supported (yet)?")
      case ForallRole(r, c) => //throw new Exception("translation of concept not supported (yet)?")
        throw new Exception("translation of concept not supported (yet)?")
      case ExistsRole(r, c) => //throw new Exception("translation of concept not supported (yet)?")
        throw new Exception("translation of concept not supported (yet)?")
      case Nominal(v) => v match {
        case Const(n) => factory.getOWLObjectOneOf(factory.getOWLNamedIndividual(":" + n, pm)).toString
        case Var(n) => "ObjectOneOf(?" + n + ")"
        case Literal(n) => throw new Exception("translation of concept/literal not supported: " + concept)
      }
      case At(v, c) => throw new Exception("translation of At-concept not supported: " + concept)
      case _ => throw new Exception("translation of concept not supported: " + concept)
    }
  }

}
