package src.ship.rest

import akka.actor.ActorDSL._
import akka.actor.ActorLogging
import com.sun.net.httpserver.{HttpServer, HttpExchange, HttpHandler}
import java.io.{ByteArrayOutputStream, PrintStream, OutputStream}
import java.net.HttpURLConnection._
import java.net.{URLDecoder, InetSocketAddress}
import org.semanticweb.owlapi.apibinding.OWLManager
import org.semanticweb.owlapi.model.OWLOntologyManager
import src.ship._

/**
 * Created with IntelliJ IDEA.
 * User: dodi
 * Date: 15.01.13
 * Time: 17:13
 * RestInterface to process commands send from the BAALL App (SHIP/src/APP)
 * example to send an abox update: http://localhost:8000/SHIPRest/update?abox=situation%3aEmergency%2ctime%3aDay
 */
class WebAppHttpHandler extends HttpHandler with ABoxFormulaParser with SHIPToolParser {

  private var currentontology: Ontology = EmptyOntology

  def setOntology(o: Ontology) {
    currentontology = o
  }

  def parseQuery(httpExchange: HttpExchange): Map[String, String] = {
    var query = httpExchange.getRequestURI.getQuery
    if (query == null) query = ""
    val pairs: Array[String] = query.split("&")
    var res: Map[String, String] = Map[String, String]()
    for (pair <- pairs) {
      val assignment: Array[String] = pair.split("=")
      if (assignment.length > 1)
        res += (assignment.apply(0) -> assignment.apply(1))
    }
    res
  }

  def sendResponse(httpExchange: HttpExchange, httpCode: Int, response: String) {
    httpExchange.sendResponseHeaders(httpCode, response.length())
    val os: OutputStream = httpExchange.getResponseBody
    os.write(response.getBytes)
    os.flush()
    os.close()
  }

  def sendOk(httpExchange: HttpExchange) {
    sendResponse(httpExchange, HTTP_OK, "This is the response")
  }

  def handle(httpExchange: HttpExchange) {
    val path: String = httpExchange.getRequestURI.getPath.substring("/SHIPRest/".length)
    if (path != "infos")
      log.info("http handle: '" + path + "'")
    // else print(".")
    path match {
      case "update" => handleUpdate(httpExchange)
      case "query" => handleQuery(httpExchange)
      case "run" => handleRun(httpExchange)
      case "cmd" => handleCmd(httpExchange)
      case "infos" => handleInfos(httpExchange)
      case "ontology" => handleOntology(httpExchange)
      case "getdiff" => handleGetDiff(httpExchange)
      case _ =>
        log.warning("cannot process path " + path)
        sendOk(httpExchange)
    }
  }

  def handleOntology(httpExchange: HttpExchange) {
    val manager: OWLOntologyManager = OWLManager.createOWLOntologyManager()
    val pm = Builtin.prefixManager
    val owlont = Ontology2OWL.convert(currentontology, pm, manager)


    val outputStream = new ByteArrayOutputStream()
    val printstream = new PrintStream(outputStream)

    manager.saveOntology(owlont, printstream)
    printstream.flush()
    val response: String = new String(outputStream.toByteArray)

    log.info("Sending:" + response)
    sendResponse(httpExchange, HTTP_ACCEPTED, response)
  }

  def handleRun(httpExchange: HttpExchange) {
    val queryMap: Map[String, String] = parseQuery(httpExchange)
    val program = URLDecoder.decode(queryMap.get("prog").get, "UTF8")
    //println("parse "+program)
    parseAll(dlprog, program) match {
      case Success(res: DLProgram, _) =>
        log.info(res.ground.toString)
        SHIPTool.ref ! Run(res.ground)
      case v => log.error("Could not parse run " + program)
    }
    sendOk(httpExchange)
  }

  def handleQuery(httpExchange: HttpExchange) {
    val queryMap: Map[String, String] = parseQuery(httpExchange)
    val queryString = URLDecoder.decode(queryMap.get("pattern").get, "UTF8")
    val substs = currentontology.queryGround(queryString, groundFormulas = true)

    def encodeSubst(subst: Substitution): String =
      "{" + subst.map.map(p => "\"" + p._1 + "\":\"" + p._2 + "\"").mkString(",\n") + "}"

    var response: String = "[" + (substs map encodeSubst).mkString(",") + "]"

    response =
      queryMap.get("callback") match {
        case None => response
        case Some(cb) => cb + "({" + response + "})"
      }

    log.info("Response to " + queryString + "\n" + substs + "\n" + response)
    httpExchange.getResponseHeaders.add("Content-type", "text/json")
    sendResponse(httpExchange, HTTP_ACCEPTED, response)

  }

  var latestgetdiffontology: Ontology = EmptyOntology

  def handleGetDiff(httpExchange: HttpExchange) {
    val diff = currentontology.computeDiff(latestgetdiffontology)

    latestgetdiffontology = currentontology

    def encode(p: (String, List[ABoxFormula])): String = {
      val aboxformulas = "[" + p._2.map(x => "\"" + x + "\"").mkString(",") + "]"
      "\"" + p._1 + "\":" + aboxformulas
    }
    var response: String = encode("deleted", diff._1) + "," + encode("added", diff._2)
    response = "{" + response + "}"
    httpExchange.getResponseHeaders.add("Content-type", "text/json")
    sendResponse(httpExchange, HTTP_ACCEPTED, response)
  }

  def handleInfos(httpExchange: HttpExchange) {
    val queryMap: Map[String, String] = parseQuery(httpExchange)
    var status = Map[String, String](
      "Paul" -> "Hide",
      "Eve" -> "Hide",
      "Rolland" -> "Hide",
      "Rolland2" -> "Hide",
      "PaulGoal" -> "None",
      "EveGoal" -> "None",
      "RollandAssignedTo" -> "None",
      "Rolland2AssignedTo" -> "None")
    val atRole = NamedRole("at")
    val wccarriesRole = NamedRole("wheelchair_carries")
    val goalsubject = NamedRole("goal_subject")
    val goaltarget = NamedRole("goal_target")
    val goalwc = NamedRole("goal_wheelchair")
    val it = status.keysIterator
    while (it.hasNext) {
      val name = it.next()
      currentontology.query(atRole.roleAssert(Const(name), Var("pos"))) match {
        case Nil =>
        case s :: _ => status = status updated(name, s(Var("pos")).asInstanceOf[Const].name)
      }
      currentontology.query(wccarriesRole.roleAssert(Var("wc"), Const(name))) match {
        case Nil =>
        case s :: _ => status = status updated(name, s(Var("wc")).asInstanceOf[Const].name)
      }

      if (name == "Paul" || name == "Eve")
        currentontology.query(List(goalsubject.roleAssert(Var("r"), Const(name)), goaltarget.roleAssert(Var("r"), Var("goal")))) match {
          case Nil =>
          case s :: _ => status = status updated(name + "Goal", s(Var("goal")).asInstanceOf[Const].name)
        }
      if (name == "Rolland" || name == "Rolland2")
        currentontology.query(List(goalwc.roleAssert(Var("r"), Const(name)), goalsubject.roleAssert(Var("r"), Var("person")))) match {
          case Nil =>
          case s :: _ => status = status updated(name + "AssignedTo", s(Var("person")).asInstanceOf[Const].name)
        }
    }
    def encode(p: (String, String)): String = "\"" + p._1 + "\":\"" + p._2 + "\""
    var response: String = status.map(encode).mkString(",")
    response = queryMap.getOrElse("callback", "showInfos") + "({" + response + "})"
    httpExchange.getResponseHeaders.add("Content-type", "text/json")
    sendResponse(httpExchange, HTTP_ACCEPTED, response)
  }

  def handleCmd(httpExchange: HttpExchange) {
    val queryMap: Map[String, String] = parseQuery(httpExchange)
    val shipcmd = queryMap.get("arg") match {
      case Some(c) => URLDecoder.decode(c, "UTF8")
      case None =>
        log.warning("Missing cmd?arg=")
        ""
    }
    parseAll(cmdListParser, shipcmd) match {
      case Success(res, _) =>
        //log.warning("WebApp handleCmd " + res.length + " " + res)
        processParseCmd(res)
      case v => log.warning("Error parsing: " + shipcmd + "\n" + v)
    }
    sendOk(httpExchange)
  }


  def handleUpdate(httpExchange: HttpExchange) {
    val queryMap: Map[String, String] = parseQuery(httpExchange)
    val aboxformulasdecoded = URLDecoder.decode(queryMap.get("abox").get, "UTF8")
    log.info("parse " + aboxformulasdecoded)
    val aboxupdates = parseABoxs(aboxformulasdecoded)
    log.info(aboxupdates.toString())
    try
      SHIPTool.ref ! Update(aboxupdates map (_.ground), List[Const]())
    catch {
      case e: Throwable => log.error(e, "got exception...")
    }
    sendOk(httpExchange)
  }
}

class BAALLApp(port: Int) extends ActWithStash with ActorLogging {
  var server: HttpServer = null
  val webapphandler: WebAppHttpHandler = new WebAppHttpHandler()

  override def preStart() {
    log.info("BAALLApp started *** ")
    try {
      val sa = new InetSocketAddress(port)
      server = HttpServer.create(sa, 0)
      server.createContext("/SHIPRest", webapphandler)
      // server.setExecutor(null); // creates a default executor
      server.start()
    } catch {
      case e: Throwable => log.error(e, "BAALLApp could not be started")
    }
  }

  override def receive = {
    case CurrentOntology(o) => webapphandler.setOntology(o)
    case _ =>
  }

  def stop() {
    log.info("Stopping BAALLApp")
    if (server != null) {
      server.stop(1)
      server = null
    }
  }
}
