package src.test

import aterm.{AFun, ATermAppl, ATerm}
import com.clarkparsia.pellet.owlapiv3.{PelletReasonerFactory, PelletReasoner}
import java.io.{OutputStreamWriter, File, FileReader}
import java.lang.String
import org.mindswap.pellet.utils.ATermUtils
import org.semanticweb.owlapi.apibinding.OWLManager
import org.semanticweb.owlapi.model._
import scala.collection.mutable
import src.ship._

/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 7/16/12
 * Time: 12:06 PM
 * This class provides interface functions to access the reasoning facilities provided by pellet as well as functions to
 * update the ontology that is managed by pellet. For debugging, the function debugOntology is provided
 */

class PelletInterface(var reasoner: PelletReasoner) {
  /** hashmap to store all defined concepts **/
  var exconcepts: mutable.HashMap[String, ATermAppl] = mutable.HashMap()
  /** hashmap to store all defined roles **/
  var exroles: mutable.HashMap[String, ATermAppl] = mutable.HashMap()
  /** namespace prefix **/
  val outwriter = new OutputStreamWriter(System.out)

  /**
   * translates a concept to an Aterm that can be digested by Pellet
   * in the case that a subterm (concept) does not exist, it is added to the ontology as new concept
   */
  def role2Pellet(role: Role): ATerm = {
    role match {
      case NamedRole(n) =>
        exroles.get(n) match {
          case Some(x) => x
          case None => addRole(n)
        }
      case NotRole(r) => ATermUtils.makeNot(role2Pellet(r).asInstanceOf[ATermAppl])
      case InverseRole(r) => ATermUtils.makeInv(role2Pellet(r).asInstanceOf[ATermAppl])
      case _ => throw new Exception("translation of role to pellet not supported: " + role)
    }
  }

  /**
   * translates a concept to an Aterm that can be digested by Pellet
   * in the case that a subterm (concept) does not exist, it is added to the ontology as new concept
   */
  def concept2Pellet(concept: Concept): ATerm = {
    concept match {
      case ConceptConstant(s) =>
        exconcepts.get(s) match {
          case Some(x) => x
          case None => addConcept(s)
        }
      case NotConcept(c) => ATermUtils.makeNot(concept2Pellet(c))
      case AndConcept(c1, c2) => ATermUtils.makeAnd(concept2Pellet(c1), concept2Pellet(c2))
      case OrConcept(c1, c2) => ATermUtils.makeOr(ATermUtils.makeList(concept2Pellet(c1), ATermUtils.makeList(concept2Pellet(c2))))
      case ForallRole(r, c) => ATermUtils.makeAllValues(role2Pellet(r), concept2Pellet(c))
      case ExistsRole(r, c) => ATermUtils.makeSomeValues(role2Pellet(r), concept2Pellet(c))
      case _ => throw new Exception("translation of concept to pellet not supported: " + concept)
    }
  }


  /**
   * adds the specified tbox formula to pellets database
   */
  def addTBoxFormula(tbox: TBoxFormula) {
    tbox match {
      case ConceptEquality(ConceptConstant(s), rhs) =>
        reasoner.getKB.addEquivalentClass(addConcept(s), concept2Pellet(rhs).asInstanceOf[ATermAppl])
      case ConceptInclusion(ConceptConstant(sub), sup) =>
        addSubConcept(sub, concept2Pellet(sup).asInstanceOf[ATermAppl])
      case _ =>
        throw new Exception("Cannot add TBOX " + tbox + ". Is LHS atomic?")
    }
  }

  /**
   * adds the subconcept to the database and defines it to be a subclass of superconcept
   */
  def addSubConcept(sub: String, superConcept: String) {
    val subConcept: ATermAppl = addConcept(sub)
    val superConceptTerm: ATermAppl = ATermUtils.makeTermAppl(superConcept)
    reasoner.getKB.addSubClass(subConcept, superConceptTerm)
  }

  /**
   * adds the subconcept to the database and defines it to be a subclass of superconcept
   */

  def addSubConcept(sub: String, superConcept: ATermAppl) {
    val subConcept: ATermAppl = addConcept(sub)
    reasoner.getKB.addSubClass(subConcept, superConcept)
    addSubConcept2Java(sub, superConcept, outwriter)
  }

  /**
   * adds the role s to the database
   */
  def addRole(s: String): ATermAppl = {
    val newrole = ATermUtils.makeTermAppl(s)
    exroles.put(s, newrole)
    reasoner.getKB.addObjectProperty(newrole)
    newrole
  }


  /**
   * adds the specified rbox formula to pellets database
   */
  def addRBoxFormula(rBoxFormula: RBoxFormula) {
    rBoxFormula match {
      case RoleEquality(x :: List(), role) =>
        reasoner.getKB.addEquivalentProperty(role2Pellet(role).asInstanceOf[ATermAppl], role2Pellet(x).asInstanceOf[ATermAppl])
      case RoleEquality(rolecomposition, role) => println(rolecomposition.length + " " + rolecomposition)
      /* case RoleInclusion(rolecomposition, role) => throw new Exception("translation of concept not supported (yet)?")
      case RoleSym(role) => throw new Exception("translation of concept not supported (yet)?")
      case RoleTra(role) => throw new Exception("translation of concept not supported (yet)?")
      case RoleRef(role) => throw new Exception("translation of concept not supported (yet)?")
      case RoleIrr(role) => throw new Exception("translation of concept not supported (yet)?")
      case RoleDis(role) => throw new Exception("translation of concept not supported (yet)?") */
      case _ => throw new Exception("translation of concept not supported (yet)?" + rBoxFormula + " " + rBoxFormula.getClass)
    }
  }

  /**
   * adds the specified abox formula to pellets database
   */
  def addAboxFormula(aBoxFormula: ABoxFormula) {
    aBoxFormula match {
      case ConceptAssertion(Var(name), concept) => throw new Exception("translation of concept not supported (yet)?")
      case ConceptAssertion(Const(name), concept) => addIndividuum(name, concept)
      case RoleAssertion(Const(src), Const(trg), role) =>
        reasoner.getKB.addObjectProperty(ATermUtils.makePropAtom(role2Pellet(role).asInstanceOf[ATermAppl],
          ATermUtils.makeTermAppl(src), ATermUtils.makeTermAppl(trg)))
        addRoleAssertion2Java(src, trg, role, outwriter)
      case _ => throw new Exception("cannot process ")
    }
  }

  /**
   * processes the specified ontology and adds all abox, tbox, and rbox formulas to the database
   */
  def processOntology(ontology: Ontology) {
    outwriter.write("//RBox\n")
    for (rbox: RBoxFormula <- ontology.rbox) {
      outwriter.write("//add " + rbox + "\n")
      addRBoxFormula(rbox)
    }

    outwriter.write("\n//TBox\n")
    for (tbox: TBoxFormula <- ontology.tbox) {
      outwriter.write("//add " + tbox + "\n")
      addTBoxFormula(tbox)
    }

    outwriter.write("\n//ABox\n")
    for (abox: ABoxFormula <- ontology.abox) {
      outwriter.write("//add " + abox + "\n")
      addAboxFormula(abox)
    }
    outwriter.flush()
  }

  /**
   * adds a new individuum with name name of type concept to pellets databse
   */
  def addIndividuum(name: String, concept: String) {
    val conceptTerm: ATermAppl = ATermUtils.makeTermAppl(concept)
    val individualTerm: ATermAppl = ATermUtils.makeTermAppl(name)
    reasoner.getKB.addIndividual(individualTerm)
    reasoner.getKB.addType(individualTerm, conceptTerm)

    addIndividuum2Java(name, concept, outwriter)
  }

  /**
   * adds a new individuum with name name of type concept to pellets databse
   */
  def addIndividuum(name: String, concept: Concept) {
    val conceptTerm: ATermAppl = concept2Pellet(concept).asInstanceOf[ATermAppl]
    val individualTerm: ATermAppl = ATermUtils.makeTermAppl(name)
    reasoner.getKB.addIndividual(individualTerm)
    reasoner.getKB.addType(individualTerm, conceptTerm)
    addIndividuum2Java(name, concept, outwriter)
  }

  /**
   * adds an atomic concept with name name to pellets database
   */
  def addConcept(name: String): ATermAppl = {
    val concept = ATermUtils.makeTermAppl(name)
    reasoner.getKB.addClass(concept)
    exconcepts.put(name, concept)

    addConcept2Java(name, outwriter)
    concept
  }

  /**
   * prints all abox formulas of pellets database
   */
  private def printABox() {
    System.out.println("=== ABOX ===")
    import scala.collection.JavaConversions._
    for (n <- reasoner.getKB.getABox.getNodes) {
      System.out.println(n + " " + n.getTypes)
    }
    System.out.println("======")
  }

  /**
   * prints all tbox formulas of pellets database
   */
  private def printTBox() {
    System.out.println("=== TBOX ===")
    import scala.collection.JavaConversions._
    for (term <- reasoner.getKB.getTBox.getAxioms) {
      System.out.println(term)
    }
    System.out.println("======")
  }

  /**
   * prints all rbox formulas of pellets database
   */
  private def printRBox() {
    System.out.println("=== RBOX ===")
    import scala.collection.JavaConversions._
    for (term <- reasoner.getKB.getRBox.getRoleNames) {
      System.out.println(term)
    }
    System.out.println("======")
  }


  /**
   * this function prints important information about the ontology and is supposed to be used for debugging
   */
  def debugOntology() {
    printABox()
    printTBox()
    printRBox()
  }

  def queryInstances(s: String, direct: Boolean) {
    exconcepts.get(s) match {
      case Some(t) => reasoner.getKB.getInstances(t, direct)
      case _ => println(exconcepts)
        throw new Exception("Concept " + s + " not registered in database")
    }
  }

  /**
   * generates the java code that adds an individuum name:concept
   */
  def addIndividuum2Java(name: String, concept: String, out: OutputStreamWriter) {
    out.write("reasoner.getKB().addIndividual( ATermUtils.makeTermAppl(\"" + name + "\"));\n")
    out.write("reasoner.getKB().addType(ATermUtils.makeTermAppl(\"" + name + "\"), ATermUtils.makeTermAppl(\"" + concept + "\"));\n")
  }

  /**
   * generates the java code that adds an individuum name:concept
   */
  def addIndividuum2Java(name: String, concept: Concept, out: OutputStreamWriter) {
    out.write("reasoner.getKB().addIndividual( ATermUtils.makeTermAppl(\"" + name + "\"));\n")
    out.write("reasoner.getKB().addType(ATermUtils.makeTermAppl(\"" + name + "\"), ATermUtils.makeTermAppl(\"" + concept + "\"));\n")
  }

  /**
   * generates the java code that adds a subclass
   */
  def addSubConcept2Java(sub: String, superConcept: ATermAppl, out: OutputStreamWriter) {
    out.write("reasoner.getKB().addSubClass(ATermUtils.makeTermAppl(\"" + sub + "\"), ")
    aterm2Java(superConcept, outwriter)
    out.write(");\n")
  }

  /**
   * generates java representation of an aterm
   */
  def aterm2Java(term: ATerm, out: OutputStreamWriter) {
    term match {
      case appl: ATermAppl =>
        if (appl.getArity == 0) out.write("ATermUtils.makeTermAppl(\"" + appl.getName + "\")")
        else out.write("ATermUtils.makeTermAppl(factory.makeAFun(\"" + appl.getName + "\"," + appl.getArity + ")")
      case _: AFun =>
        val appl = term.asInstanceOf[ATermAppl]
        out.write("factory.makeAFun(\"" + appl.getName + "\"," + appl.getArity + ")")
      case _ => out.write("cannot create constructor for " + term)
    }
  }

  /**
   * generates the java representation for an assertion
   */
  def addRoleAssertion2Java(src: String, trg: String, role: Role, out: OutputStreamWriter) {
    out.write("reasoner.getKB().addObjectProperty(ATermUtils.makePropAtom(")
    aterm2Java(role2Pellet(role), out)
    out.write(",")
    out.write("ATermUtils.makeTermAppl(\"" + src + "\"), ATermUtils.makeTermAppl(\"" + trg + "\")));\n")
  }

  /**
   * generates the java representation to add a concept
   */
  def addConcept2Java(name: String, out: OutputStreamWriter) {
    out.write("reasoner.getKB().addClass(ATermUtils.makeTermAppl(\"" + name + "\"));\n")
  }
}


object PelletTest extends OntologyParser {
  def main(args: Array[String]) {
    var manager = OWLManager.createOWLOntologyManager()
    var pelletontology = manager.createOntology()
    var preasoner = new PelletInterface(PelletReasonerFactory.getInstance().createReasoner(pelletontology))

    /*preasoner.addConcept("Position")
    preasoner.addConcept("WheelChair")
    preasoner.addSubConcept("TransportAssistant", "WheelChair")*/

    parseAll(ontology, new FileReader(new File("examples/baall/transport.ont"))) match {
      case Success(s, in) => preasoner.processOntology(s)
        preasoner.debugOntology()
        preasoner.reasoner.getKB.realize()
        preasoner.reasoner.getKB.printClassTree()
        preasoner.reasoner.prepareReasoner()
        val factory: OWLDataFactory = manager.getOWLDataFactory
        val accessibleclass = factory.getOWLClass(IRI.create("a:Position"))

        println(accessibleclass)
        println(preasoner.reasoner.getInstances(accessibleclass, false))


        val iter = preasoner.reasoner.getKB.getClasses.iterator()
        while (iter.hasNext) {
          val c: ATermAppl = iter.next()
          println("instances of " + c + " " + preasoner.reasoner.getKB.getInstances(c, false))
        }

        // this is for getting the real class..
        println(preasoner.reasoner.getInstances(preasoner.reasoner.getTopClassNode.getRepresentativeElement, false))
        println(preasoner.reasoner.getInstances(factory.getOWLClass(IRI.create("owl:Thing")), false))
        println(preasoner.reasoner.getInstances(factory.getOWLClass(IRI.create("Thing")), false))

      //PrefixManager pm = new DefaultPrefixManager("http://www.people.lu.unisi.ch/fornaran/ontologies/OntologyOBL.owl#");
      //OWLClass Cancelled = factory.getOWLClass(":Cancelled", pm);

      case NoSuccess(msg, _) => println("failure: " + msg)
    }
  }
}
